from typing import TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.hardware.hardware_base import InsufficientCapacityError

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern


class EdgeCPUComponent(EdgeComponent):
    compatible_root_units = [u.cpu_core]
    default_values = {
        "carbon_footprint_fabrication": SourceValue(20 * u.kg),
        "power": SourceValue(15 * u.W),
        "lifespan": SourceValue(6 * u.year),
        "idle_power": SourceValue(3 * u.W),
        "compute": SourceValue(4 * u.cpu_core),
        "base_compute_consumption": SourceValue(0.1 * u.cpu_core),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 compute: ExplainableQuantity, base_compute_consumption: ExplainableQuantity):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, idle_power)
        self.compute = compute.set_label(f"Compute of {self.name}")
        self.base_compute_consumption = base_compute_consumption.set_label(f"Base compute consumption of {self.name}")

        self.available_compute_per_instance = EmptyExplainableObject()
        self.unitary_hourly_compute_need_per_usage_pattern = ExplainableObjectDict()

    @property
    def calculated_attributes(self):
        return (["available_compute_per_instance", "unitary_hourly_compute_need_per_usage_pattern"]
                + super().calculated_attributes)

    def update_available_compute_per_instance(self):
        available_compute_per_instance = (self.compute - self.base_compute_consumption)

        if available_compute_per_instance < SourceValue(0 * u.cpu_core):
            raise InsufficientCapacityError(self, "compute", self.compute, self.base_compute_consumption)

        self.available_compute_per_instance = available_compute_per_instance.set_label(
            f"Available compute per {self.name} instance")

    def update_dict_element_in_unitary_hourly_compute_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_compute_need = sum(
            [need.unitary_hourly_need_per_usage_pattern[usage_pattern]
             for need in self.recurrent_edge_component_needs if usage_pattern in need.edge_usage_patterns],
            start=EmptyExplainableObject())

        if not isinstance(unitary_hourly_compute_need, EmptyExplainableObject):
            max_compute_need = unitary_hourly_compute_need.max().to(u.cpu_core)
            if max_compute_need > self.available_compute_per_instance:
                raise InsufficientCapacityError(self, "compute", self.available_compute_per_instance, max_compute_need)

        self.unitary_hourly_compute_need_per_usage_pattern[usage_pattern] = unitary_hourly_compute_need.to(
            u.cpu_core).set_label(f"{self.name} hourly compute need for {usage_pattern.name}").generate_explainable_object_with_logical_dependency(
            self.available_compute_per_instance)

    def update_unitary_hourly_compute_need_per_usage_pattern(self):
        self.unitary_hourly_compute_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_compute_need_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_power_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        if usage_pattern in self.unitary_hourly_compute_need_per_usage_pattern:
            compute_need = self.unitary_hourly_compute_need_per_usage_pattern[usage_pattern]
        else:
            compute_need = EmptyExplainableObject()

        if isinstance(compute_need, EmptyExplainableObject):
            unitary_power = self.idle_power
        else:
            unitary_compute_workload = (compute_need + self.base_compute_consumption) / self.compute
            unitary_power = self.idle_power + (self.power - self.idle_power) * unitary_compute_workload

        self.unitary_power_per_usage_pattern[usage_pattern] = unitary_power.set_label(
            f"{self.name} unitary power for {usage_pattern.name}")

    def update_unitary_power_per_usage_pattern(self):
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_power_per_usage_pattern(usage_pattern)



import math
from copy import copy
from typing import List, TYPE_CHECKING, Optional

import numpy as np
from pint import Quantity

from efootprint.constants.sources import Sources
from efootprint.core.hardware.infra_hardware import InfraHardware
from efootprint.core.hardware.hardware_base import InsufficientCapacityError
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.job import JobBase
    from efootprint.core.hardware.server_base import ServerBase


class NegativeCumulativeStorageNeedError(Exception):
    def __init__(self, storage_obj: "Storage", cumulative_quantity: Quantity):
        self.storage_obj = storage_obj
        self.cumulative_quantity = cumulative_quantity
        self.jobs_that_delete_data = [job for job in self.storage_obj.jobs if job.data_stored.magnitude < 0]

        job_msg_part = [f"name: {job.name} - value: {job.data_stored}" for job in self.jobs_that_delete_data]

        message = (
            f"In Storage object {self.storage_obj.name}, negative cumulative storage need detected: "
            f"{np.min(cumulative_quantity):~P}. Please verify your jobs that delete data: {job_msg_part} "
            f"or increase the base_storage_need value, currently set to {self.storage_obj.base_storage_need.value}")
        super().__init__(message)


class Storage(InfraHardware):
    default_values =  {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(160 * u.kg / u.TB),
            "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB),
            "lifespan": SourceValue(6 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB),
            "data_replication_factor": SourceValue(3 * u.dimensionless),
            "base_storage_need": SourceValue(0 * u.TB),
            "data_storage_duration": SourceValue(5 * u.year)
        }

    @classmethod
    def ssd(cls, name="Default SSD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                160 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(6 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "data_replication_factor": SourceValue(3 * u.dimensionless),
            "base_storage_need": SourceValue(0 * u.TB),
            "data_storage_duration": SourceValue(5 * u.year)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def hdd(cls, name="Default HDD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                20 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(4.2 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(4 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "data_replication_factor": SourceValue(3 * u.dimensionless),
            "base_storage_need": SourceValue(0 * u.TB),
            "data_storage_duration": SourceValue(5 * u.year)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def archetypes(cls):
        return [cls.ssd, cls.hdd]

    def __init__(self, name: str, storage_capacity: ExplainableQuantity,
                 carbon_footprint_fabrication_per_storage_capacity: ExplainableQuantity,
                 power_per_storage_capacity: ExplainableQuantity, idle_power: ExplainableQuantity,
                 data_replication_factor: ExplainableQuantity, data_storage_duration: ExplainableQuantity,
                 base_storage_need: ExplainableQuantity, lifespan: ExplainableQuantity,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject = None):
        super().__init__(
            name, carbon_footprint_fabrication=SourceValue(0 * u.kg), power=SourceValue(0 * u.W), lifespan=lifespan)
        self.carbon_footprint_fabrication_per_storage_capacity = (carbon_footprint_fabrication_per_storage_capacity
        .set_label(f"Fabrication carbon footprint of {self.name} per storage capacity"))
        self.power_per_storage_capacity = power_per_storage_capacity.set_label(
            f"Power of {self.name} per storage capacity")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.storage_capacity = storage_capacity.set_label(f"Storage capacity of {self.name}")
        self.data_replication_factor = data_replication_factor.set_label(f"Data replication factor of {self.name}")
        self.data_storage_duration = data_storage_duration.set_label(f"Data storage duration of {self.name}")
        self.base_storage_need = base_storage_need.set_label(f"{self.name} initial storage need")
        self.fixed_nb_of_instances = (fixed_nb_of_instances or EmptyExplainableObject()).set_label(
            f"User defined number of {self.name} instances").to(u.concurrent)
        self.storage_delta = EmptyExplainableObject()
        self.full_cumulative_storage_need = EmptyExplainableObject()
        self.nb_of_active_instances = EmptyExplainableObject()
        self.storage_needed = EmptyExplainableObject()
        self.storage_freed = EmptyExplainableObject()
        self.automatic_storage_dumps_after_storage_duration = EmptyExplainableObject()

    @property
    def server(self) -> Optional["ServerBase"]:
        if self.modeling_obj_containers:
            if len(self.modeling_obj_containers) > 1:
                raise PermissionError(
                    f"Storage object can only be associated with one server object but {self.name} is associated "
                    f"with {[mod_obj.name for mod_obj in self.modeling_obj_containers]}")
            return self.modeling_obj_containers[0]
        else:
            return None

    @property
    def calculated_attributes(self):
        return [
            "carbon_footprint_fabrication", "power", "storage_needed", "storage_freed",
            "automatic_storage_dumps_after_storage_duration", "storage_delta", "full_cumulative_storage_need",
            "raw_nb_of_instances", "nb_of_instances", "nb_of_active_instances", "instances_fabrication_footprint",
            "instances_energy", "energy_footprint"]

    @property
    def jobs(self) -> List["JobBase"]:
        server = self.server
        if server is not None:
            return server.jobs
        return []

    @property
    def power_usage_effectiveness(self):
        if self.server is not None:
            return self.server.power_usage_effectiveness
        else:
            return EmptyExplainableObject()

    @property
    def average_carbon_intensity(self):
        if self.server is not None:
            return self.server.average_carbon_intensity
        else:
            return EmptyExplainableObject()

    def update_carbon_footprint_fabrication(self):
        self.carbon_footprint_fabrication = (
                self.carbon_footprint_fabrication_per_storage_capacity * self.storage_capacity).set_label(
            f"Carbon footprint of {self.name}")

    def update_power(self):
        self.power = (self.power_per_storage_capacity * self.storage_capacity).set_label(f"Power of {self.name}")

    def update_storage_needed(self):
        storage_needed = EmptyExplainableObject()

        for job in self.jobs:
            if job.data_stored.magnitude >= 0:
                storage_needed += job.hourly_data_stored_across_usage_patterns
            else:
                # Necessary to keep relationship in calculation graph.
                # For example if the data_stored attribute of a job went from negative to positive then the
                # storage_needed needs to be recomputed but wouldn’t have been initially logically linked to the job
                # with the above if condition.
                storage_needed = storage_needed.generate_explainable_object_with_logical_dependency(
                    job.hourly_data_stored_across_usage_patterns)

        storage_needed *= self.data_replication_factor

        self.storage_needed = storage_needed.to(u.TB).set_label(f"Hourly {self.name} storage need")

    def update_storage_freed(self):
        storage_freed = EmptyExplainableObject()

        for job in self.jobs:
            if job.data_stored.magnitude < 0:
                storage_freed += job.hourly_data_stored_across_usage_patterns
            else:
                storage_freed = storage_freed.generate_explainable_object_with_logical_dependency(
                    job.hourly_data_stored_across_usage_patterns)

        storage_freed *= self.data_replication_factor

        self.storage_freed = storage_freed.to(u.TB).set_label(f"Hourly {self.name} storage freed")

    def update_automatic_storage_dumps_after_storage_duration(self):
        if isinstance(self.storage_needed, EmptyExplainableObject):
            automatic_storage_dumps_after_storage_duration = EmptyExplainableObject(left_parent=self.storage_needed)
        else:
            storage_needs_nb_of_hours = len(self.storage_needed.value)
            if storage_needs_nb_of_hours == 0:
                automatic_storage_dumps_after_storage_duration_np = Quantity(
                    np.zeros(storage_needs_nb_of_hours + 1, dtype=np.float32),
                    self.storage_needed.units)
            else:
                # Use copy not to modify data storage duration unit in place
                storage_duration_in_hours = math.ceil(copy(self.data_storage_duration.to(u.hour)).magnitude)
                automatic_storage_dumps_after_storage_duration_np = - np.pad(
                    self.storage_needed.value, (storage_duration_in_hours, 0), constant_values=np.float32(0)
                )[:storage_needs_nb_of_hours]

            automatic_storage_dumps_after_storage_duration = ExplainableHourlyQuantities(
                automatic_storage_dumps_after_storage_duration_np, self.storage_needed.start_date,
                label=f"Storage dumps for {self.name}",
                left_parent=self.storage_needed,
                right_parent=self.data_storage_duration, operator="shift by storage duration and negate")

        self.automatic_storage_dumps_after_storage_duration = automatic_storage_dumps_after_storage_duration

    def update_storage_delta(self):
        storage_delta = (self.storage_needed + self.storage_freed
                         + self.automatic_storage_dumps_after_storage_duration)

        self.storage_delta = storage_delta.set_label(f"Hourly storage delta for {self.name}")

    def update_full_cumulative_storage_need(self):
        if isinstance(self.storage_delta, EmptyExplainableObject):
            self.full_cumulative_storage_need = EmptyExplainableObject(left_parent=self.storage_delta)
        else:
            delta_array = np.copy(self.storage_delta.value.magnitude)
            delta_unit = self.storage_delta.value.units

            # Add base storage need to first hour
            delta_array[0] += self.base_storage_need.value.to(delta_unit).magnitude

            # Compute cumulative storage
            cumulative_array = np.cumsum(delta_array, dtype=np.float32)
            cumulative_quantity = Quantity(cumulative_array, delta_unit)

            if np.min(cumulative_quantity.magnitude) < 0:
                raise NegativeCumulativeStorageNeedError(self, cumulative_quantity)

            self.full_cumulative_storage_need = ExplainableHourlyQuantities(
                cumulative_quantity,
                start_date=self.storage_delta.start_date,
                label=f"Full cumulative storage need for {self.name}",
                left_parent=self.storage_delta,
                right_parent=self.base_storage_need,
                operator="cumulative sum of storage delta with initial storage need"
            )

    def update_raw_nb_of_instances(self):
        raw_nb_of_instances = (self.full_cumulative_storage_need / self.storage_capacity).to(u.concurrent)

        self.raw_nb_of_instances = raw_nb_of_instances.set_label(f"Hourly raw number of instances for {self.name}")

    def update_nb_of_instances(self):
        if isinstance(self.raw_nb_of_instances, EmptyExplainableObject):
            self.nb_of_instances = EmptyExplainableObject(left_parent=self.raw_nb_of_instances)
            return

        ceiled_nb_of_instances = self.raw_nb_of_instances.ceil()

        if not isinstance(self.fixed_nb_of_instances, EmptyExplainableObject):
            max_nb_of_instances = ceiled_nb_of_instances.max()
            if max_nb_of_instances > self.fixed_nb_of_instances:
                raise InsufficientCapacityError(
                    self, "number of instances", self.fixed_nb_of_instances, max_nb_of_instances)
            else:
                fixed_nb_of_instances_quantity = Quantity(
                    np.full(
                        len(self.raw_nb_of_instances),
                        np.float32(self.fixed_nb_of_instances.to(u.concurrent).magnitude)
                    ), u.concurrent)
                fixed_nb_of_instances = ExplainableHourlyQuantities(
                    fixed_nb_of_instances_quantity, self.raw_nb_of_instances.start_date,"Nb of instances",
                    left_parent=self.raw_nb_of_instances, right_parent=self.fixed_nb_of_instances)
            self.nb_of_instances = fixed_nb_of_instances.set_label(f"Hourly fixed number of instances for {self.name}")
        else:
            nb_of_instances = ceiled_nb_of_instances.generate_explainable_object_with_logical_dependency(
                self.fixed_nb_of_instances)
            self.nb_of_instances = nb_of_instances.set_label(f"Hourly number of instances for {self.name}")

    def update_nb_of_active_instances(self):
        storage_needed = self.storage_needed
        tmp_nb_of_active_instances = (
                (storage_needed.abs().np_compared_with(self.storage_freed.abs(), "max")
                 + self.automatic_storage_dumps_after_storage_duration.abs())
                / self.storage_capacity
        ).to(u.concurrent)
        nb_of_active_instances = tmp_nb_of_active_instances.np_compared_with(self.nb_of_instances, "min")
        self.nb_of_active_instances = nb_of_active_instances.set_label(
            f"Hourly number of active instances for {self.name}")

    def update_instances_energy(self):
        nb_of_idle_instances = (self.nb_of_instances - self.nb_of_active_instances).set_label(
            f"Hourly number of idle instances for {self.name}")
        active_storage_energy = (
                self.nb_of_active_instances * self.power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly active instances energy for {self.name}")
        idle_storage_energy = (
                nb_of_idle_instances * self.idle_power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly idle instances energy for {self.name}")

        storage_energy = (active_storage_energy + idle_storage_energy)

        self.instances_energy = storage_energy.to(u.kWh).set_label(f"Storage energy for {self.name}")



from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.storage import Storage
from efootprint.core.hardware.server_base import ServerBase, ServerTypes


class Server(ServerBase):
    def _abc_marker(self):
        pass  # silent override

    default_values =  {
            "server_type": ServerTypes.autoscaling(),
            "carbon_footprint_fabrication": SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
            "power": SourceValue(300 * u.W),
            "lifespan": SourceValue(6 * u.year),
            "idle_power": SourceValue(50 * u.W),
            "ram": SourceValue(128 * u.GB_ram),
            "compute": SourceValue(24 * u.cpu_core),
            "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless),
            "average_carbon_intensity": SourceValue(400 * u.g / u.kWh),
            "utilization_rate": SourceValue(0.9 * u.dimensionless),
            "base_ram_consumption": SourceValue(0 * u.GB_ram),
            "base_compute_consumption": SourceValue(0 * u.cpu_core),
            "fixed_nb_of_instances": EmptyExplainableObject()
        }

    def __init__(self, name: str, server_type: ExplainableObject, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 ram: ExplainableQuantity, compute: ExplainableQuantity,
                 power_usage_effectiveness: ExplainableQuantity, average_carbon_intensity: ExplainableQuantity,
                 utilization_rate: ExplainableQuantity, base_ram_consumption: ExplainableQuantity,
                 base_compute_consumption: ExplainableQuantity, storage: Storage,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject = None):
        super().__init__(
            name, server_type, carbon_footprint_fabrication, power, lifespan, idle_power, ram, compute,
            power_usage_effectiveness, average_carbon_intensity, utilization_rate, base_ram_consumption,
            base_compute_consumption, storage, fixed_nb_of_instances)



from typing import List, TYPE_CHECKING

from efootprint.core.country import Country
from efootprint.core.hardware.network import Network
from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
from efootprint.core.usage.compute_nb_occurrences_in_parallel import compute_nb_avg_hourly_occurrences
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import (
    ExplainableHourlyQuantities)
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
    from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
    from efootprint.core.usage.job import JobBase


class EdgeUsagePattern(ModelingObject):
    def __init__(self, name: str, edge_usage_journey: EdgeUsageJourney, network: Network,
                 country: Country, hourly_edge_usage_journey_starts: ExplainableHourlyQuantities):
        super().__init__(name)
        self.utc_hourly_edge_usage_journey_starts = EmptyExplainableObject()
        self.nb_edge_usage_journeys_in_parallel = EmptyExplainableObject()

        self.hourly_edge_usage_journey_starts = hourly_edge_usage_journey_starts.to(u.occurrence).set_label(
            f"{self.name} hourly nb of edge device starts")
        self.edge_usage_journey = edge_usage_journey
        self.network = network
        self.country = country

    @property
    def calculated_attributes(self):
        return ["utc_hourly_edge_usage_journey_starts", "nb_edge_usage_journeys_in_parallel"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> (
            List)["RecurrentEdgeDeviceNeed | RecurrentServerNeed"]:
        return self.recurrent_edge_device_needs + self.recurrent_server_needs

    @property
    def recurrent_edge_device_needs(self) -> List["RecurrentEdgeDeviceNeed"]:
        return self.edge_usage_journey.recurrent_edge_device_needs

    @property
    def recurrent_server_needs(self) -> List["RecurrentServerNeed"]:
        return self.edge_usage_journey.recurrent_server_needs

    @property
    def jobs(self) -> List["JobBase"]:
        return self.edge_usage_journey.jobs

    def update_utc_hourly_edge_usage_journey_starts(self):
        utc_hourly_edge_usage_journey_starts = self.hourly_edge_usage_journey_starts.convert_to_utc(
            local_timezone=self.country.timezone)

        self.utc_hourly_edge_usage_journey_starts = utc_hourly_edge_usage_journey_starts.set_label(
            f"{self.name} UTC")

    def update_nb_edge_usage_journeys_in_parallel(self):
        nb_of_edge_usage_journeys_in_parallel = compute_nb_avg_hourly_occurrences(
            self.utc_hourly_edge_usage_journey_starts, self.edge_usage_journey.usage_span)

        self.nb_edge_usage_journeys_in_parallel = nb_of_edge_usage_journeys_in_parallel.to(u.concurrent).set_label(
            f"{self.name} hourly nb of edge usage journeys in parallel")



from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.hardware.hardware_base import InsufficientCapacityError

if TYPE_CHECKING:
    from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
    from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.usage.edge.edge_function import EdgeFunction
    from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed
    from efootprint.core.hardware.edge.edge_storage import EdgeStorage


class EdgeDevice(ModelingObject):
    default_values = {
        "structure_carbon_footprint_fabrication": SourceValue(50 * u.kg),
        "lifespan": SourceValue(6 * u.year)
    }

    def __init__(self, name: str, structure_carbon_footprint_fabrication: ExplainableQuantity,
                 components: List[EdgeComponent], lifespan: ExplainableQuantity):
        super().__init__(name)
        self.lifespan = lifespan.set_label(f"Lifespan of {self.name}")
        self.structure_carbon_footprint_fabrication = structure_carbon_footprint_fabrication.set_label(
            f"Structure fabrication carbon footprint of {self.name}")
        self.components = components

        self.lifespan_validation = EmptyExplainableObject()
        self.component_needs_edge_device_validation = EmptyExplainableObject()
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.instances_energy = EmptyExplainableObject()
        self.energy_footprint = EmptyExplainableObject()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @property
    def calculated_attributes(self):
        return ["lifespan_validation", "component_needs_edge_device_validation",
                "instances_fabrication_footprint_per_usage_pattern",
                "instances_energy_per_usage_pattern", "energy_footprint_per_usage_pattern",
                "instances_fabrication_footprint", "instances_energy", "energy_footprint"]

    @property
    def recurrent_edge_device_needs(self) -> List["RecurrentEdgeDeviceNeed"]:
        from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
        return [elt for elt in self.modeling_obj_containers if isinstance(elt, RecurrentEdgeDeviceNeed)]

    @property
    def recurrent_server_needs(self) -> List["RecurrentServerNeed"]:
        from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
        return [elt for elt in self.modeling_obj_containers if isinstance(elt, RecurrentServerNeed)]

    @property
    def recurrent_needs(self) -> List["RecurrentEdgeDeviceNeed | RecurrentServerNeed"]:
        return self.modeling_obj_containers

    @property
    def recurrent_edge_component_needs(self) -> List["RecurrentEdgeComponentNeed"]:
        return list(set(sum([need.recurrent_edge_component_needs
                             for need in self.recurrent_edge_device_needs], start=[])))

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return list(set(sum([need.edge_usage_journeys for need in self.recurrent_needs], start=[])))

    @property
    def edge_functions(self) -> List["EdgeFunction"]:
        return list(set(sum([need.edge_functions for need in self.recurrent_needs], start=[])))

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([need.edge_usage_patterns for need in self.recurrent_needs], start=[])))

    def _filter_component_by_type(self, component_type: type) -> List[EdgeComponent]:
        components_of_type = []
        for component in self.components:
            if isinstance(component, component_type):
                components_of_type.append(component)

        return components_of_type

    @property
    def storages(self) -> List["EdgeStorage"]:
        from efootprint.core.hardware.edge.edge_storage import EdgeStorage
        return self._filter_component_by_type(EdgeStorage)

    @property
    def cpus(self):
        from efootprint.core.hardware.edge.edge_cpu_component import EdgeCPUComponent
        return self._filter_component_by_type(EdgeCPUComponent)

    def update_lifespan_validation(self):
        result = EmptyExplainableObject().generate_explainable_object_with_logical_dependency(self.lifespan)
        for edge_usage_journey in self.edge_usage_journeys:
            if self.lifespan < edge_usage_journey.usage_span:
                raise InsufficientCapacityError(self, "lifespan", self.lifespan, edge_usage_journey.usage_span)
            result = result.generate_explainable_object_with_logical_dependency(edge_usage_journey.usage_span)
        self.lifespan_validation = result

    def update_component_needs_edge_device_validation(self):
        """Validate that all component needs point to components of this edge_device."""
        for component_need in self.recurrent_edge_component_needs:
            component_device = component_need.edge_component.edge_device
            if component_device is not None and component_device != self:
                raise ValueError(
                    f"RecurrentEdgeComponentNeed '{component_need.name}' points to component "
                    f"'{component_need.edge_component.name}' belonging to EdgeDevice '{component_device.name}', "
                    f"but RecurrentEdgeDeviceNeed '{self.name}' is linked to EdgeDevice '{self.name}'. "
                    f"All component needs must belong to the same edge device.")

        self.component_needs_edge_device_validation = EmptyExplainableObject()

    def update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(
            self, usage_pattern: "EdgeUsagePattern"):
        # Sum fabrication footprints from all components plus device structure
        structure_fabrication_intensity = self.structure_carbon_footprint_fabrication / self.lifespan
        nb_instances = usage_pattern.nb_edge_usage_journeys_in_parallel

        structure_footprint = (
            nb_instances * structure_fabrication_intensity * ExplainableQuantity(1 * u.hour, "one hour"))

        total_footprint = structure_footprint
        for component in self.components:
            if usage_pattern in component.instances_fabrication_footprint_per_usage_pattern:
                total_footprint += component.instances_fabrication_footprint_per_usage_pattern[usage_pattern]

        self.instances_fabrication_footprint_per_usage_pattern[usage_pattern] = total_footprint.to(
            u.kg).set_label(f"Hourly {self.name} instances fabrication footprint for {usage_pattern.name}")

    def update_instances_fabrication_footprint_per_usage_pattern(self):
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(usage_pattern)

    def update_dict_element_in_instances_energy_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        # Sum energy from all components
        total_energy = EmptyExplainableObject()
        for component in self.components:
            if usage_pattern in component.instances_energy_per_usage_pattern:
                total_energy += component.instances_energy_per_usage_pattern[usage_pattern]

        self.instances_energy_per_usage_pattern[usage_pattern] = total_energy.set_label(
            f"Hourly energy consumed by {self.name} instances for {usage_pattern.name}")

    def update_instances_energy_per_usage_pattern(self):
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_energy_per_usage_pattern(usage_pattern)

    def update_dict_element_in_energy_footprint_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        # Sum energy footprint from all components
        total_energy_footprint = EmptyExplainableObject()
        for component in self.components:
            if usage_pattern in component.energy_footprint_per_usage_pattern:
                total_energy_footprint += component.energy_footprint_per_usage_pattern[usage_pattern]

        self.energy_footprint_per_usage_pattern[usage_pattern] = total_energy_footprint.set_label(
            f"{self.name} energy footprint for {usage_pattern.name}")

    def update_energy_footprint_per_usage_pattern(self):
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_energy_footprint_per_usage_pattern(usage_pattern)

    def update_instances_energy(self):
        instances_energy = sum(
            self.instances_energy_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.instances_energy = instances_energy.set_label(
            f"{self.name} total energy consumed across usage patterns")

    def update_energy_footprint(self):
        energy_footprint = sum(
            self.energy_footprint_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.energy_footprint = energy_footprint.set_label(
            f"{self.name} total energy footprint across usage patterns")

    def update_instances_fabrication_footprint(self):
        instances_fabrication_footprint = sum(
            self.instances_fabrication_footprint_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.instances_fabrication_footprint = instances_fabrication_footprint.set_label(
            f"{self.name} total fabrication footprint across usage patterns")



import uuid

WIDTH = "1200px"
HEIGHT = "900px"


def set_string_max_width(input_string, max_width):
    sections = input_string.split("\n \n")
    formatted_sections = []
    for section in sections:
        lines = section.split('\n')
        formatted_lines = []
    
        for line in lines:
            words = line.split()
            current_line = []
            current_len = 0
    
            for word in words:
                if current_len + len(word) > max_width:
                    formatted_lines.append(' '.join(current_line))
                    current_line = [word]
                    current_len = len(word) + 1
                else:
                    current_line.append(word)
                    current_len += len(word) + 1
    
            if current_line:
                formatted_lines.append(' '.join(current_line))
                
        formatted_sections.append('\n'.join(formatted_lines))

    return "\n \n".join(formatted_sections)


def add_unique_id_to_mynetwork(filename):
    with open(filename, "r") as file:
        html_content = file.read()

    # Replace "mynetwork" in js scripts with "mynetwork+uuid" so that several graphs displayed on the same
    # page don’t collide with one another
    html_content = html_content.replace("mynetwork", f"mynetwork_{str(uuid.uuid4())[:6]}")

    with open(filename, "w") as file:
        file.write(html_content)



from typing import List, Dict

import matplotlib.pyplot as plt
import numpy as np

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity


class EmissionPlotter:
    def __init__(self, ax, formatted_input_dicts__old: List[Dict[str, ExplainableQuantity]],
                 formatted_input_dicts__new: List[Dict[str, ExplainableQuantity]], rounding_value: int,
                 legend_labels: List[str] = ("Electricity consumption", "Fabrication")):
        self.ax = ax
        self.formatted_input_dicts__old = formatted_input_dicts__old
        self.formatted_input_dicts__new = formatted_input_dicts__new
        self.rounding_value = rounding_value
        self.legend_labels = legend_labels
        all_elements = ["Servers", "Storage", "Network", "Devices", "EdgeComputers", "EdgeStorage"]
        elements = []
        for elt in all_elements:
            for dict in formatted_input_dicts__old + formatted_input_dicts__new:
                if elt in dict and dict[elt] != EmptyExplainableObject() and elt not in elements:
                    elements.append(elt)
        self.elements = elements
        self.index = np.arange(len(self.elements))
        self.bar_width = 0.4
        self.total_emissions_in_kg__new = self.calculate_total_emissions(formatted_input_dicts__new)
        self.total_emissions_in_kg__old = self.calculate_total_emissions(formatted_input_dicts__old)
        self.colors = ["#6372f2", "#de5f46"]
        if self.total_emissions_in_kg__new < 501:
            self.unit = "kg"
            self.dividing_number = 1
        else:
            self.unit = "tonne"
            self.dividing_number = 1000


    @staticmethod
    def calculate_total_emissions(formatted_input_dicts):
        total_emissions_in_kg = 0
        for input_dict in formatted_input_dicts:
            total_emissions_in_kg += sum(input_dict.values()).to(u.kg).magnitude
        return total_emissions_in_kg

    def get_values(self, input_dict):
        return [(input_dict.get(element, 0 * u(self.unit))).to(u(self.unit)).magnitude
                for element in self.elements]

    def plot_common_values(self, common_values, i, color):
        return self.ax.bar(self.index + i * self.bar_width, common_values, self.bar_width, color=color, alpha=1.0)

    def plot_difference_values(self, diff_values, common_values, i, color):
        return self.ax.bar(
            self.index + i * self.bar_width, diff_values, self.bar_width, bottom=common_values, color=color, alpha=0.1)

    def plot_positive_diff(self, positive_diffs, common_values, i, color):
        return self.ax.bar(
            self.index + i * self.bar_width, positive_diffs, self.bar_width, bottom=common_values, color=color,
            alpha=0.7)

    def add_annotations_and_text(self, rects_common, diffs, values_old, values_new):
        arrowprops = dict(facecolor='black', shrink=0.05, width=2, headwidth=8)

        figure_height = self.ax.get_ylim()[1]
        for rect, diff, value_old, value_new in zip(rects_common, diffs, values_old, values_new):
            if value_old != value_new:
                if diff < -1:
                    if abs(value_old - value_new) > figure_height / 10:
                        self.ax.annotate(
                            "", xy=(rect.get_x() + rect.get_width() / 2 - 0.06,
                                    min(value_old, value_new) + figure_height / 20),
                            xytext=(rect.get_x() + rect.get_width() / 2 - 0.06, max(value_old, value_new)),
                            arrowprops=arrowprops)
                    self.ax.text(
                        rect.get_x() + rect.get_width() / 2 + 0.06, (value_old + value_new) / 2, f"{diff:.0f}%",
                        ha="center", va="center")
                elif diff > 1:
                    if abs(value_old - value_new) > figure_height / 10:
                        self.ax.annotate(
                            "", xy=(rect.get_x() + rect.get_width() / 2 - 0.06, max(value_old, value_new)),
                            xytext=(rect.get_x() + rect.get_width() / 2 - 0.06, min(value_old, value_new)),
                            arrowprops=arrowprops)
                    self.ax.text(
                        rect.get_x() + rect.get_width() / 2 + 0.06, (value_old + value_new) / 2, f"+{diff:.0f}%",
                        ha="center", va="center")

            proportion = (value_new / (self.total_emissions_in_kg__new / self.dividing_number)) * 100
            self.ax.text(rect.get_x() + rect.get_width() / 2, value_new, f"{proportion:.0f}%", ha="center", va="bottom")

    def set_axes_labels(self):
        self.ax.set_xlabel("Category")
        self.ax.set_ylabel(f"{self.unit}s CO2 emissions".replace("1 ", ""))

        self.ax.set_xticks(self.index + self.bar_width / 2)
        self.ax.set_xticklabels(self.elements, rotation=0, ha="center")

        ax2 = self.ax.twinx()
        max_value = max(
            [max(input_dict.values())
             for input_dict in self.formatted_input_dicts__new + self.formatted_input_dicts__old]
        ).to(u(self.unit)).magnitude

        max_value_margin = 1.1
        ax2.set_ylim(0, 100 * max_value_margin * (max_value / (self.total_emissions_in_kg__new / self.dividing_number)))
        ax2.set_ylabel("Proportions (%)")

        self.ax.set_ylim(0, max_value_margin * max_value)

    def set_titles(self):
        rounded_total__new = round(self.total_emissions_in_kg__new / self.dividing_number, self.rounding_value)
        if self.rounding_value == 0:
            rounded_total__new = int(rounded_total__new)

        rounded_total__old = round(self.total_emissions_in_kg__old / self.dividing_number, self.rounding_value)
        if self.rounding_value == 0:
            rounded_total__old = int(rounded_total__old)

        if rounded_total__old != rounded_total__new:
            plus_sign = ""
            if rounded_total__new - rounded_total__old > 0:
                plus_sign = "+"
            subtitle_text = f"From {rounded_total__old} to {rounded_total__new} {self.unit}s of CO2 emissions " \
                            f"({plus_sign}{int(100 * (rounded_total__new - rounded_total__old) / rounded_total__old)}%)"
        else:
            subtitle_text = f"{rounded_total__new} {self.unit}s of CO2 emissions"

        self.ax.text(
            0.5, 1.1, subtitle_text,
            transform=self.ax.transAxes, fontsize=16, va="top", ha="center")

    def plot_emission_diffs(self):
        for i, (input_dict_old, input_dict_new) in enumerate(
                zip(self.formatted_input_dicts__old, self.formatted_input_dicts__new)):
            values_old = self.get_values(input_dict_old)
            values_new = self.get_values(input_dict_new)

            diffs = [(new - old) / old * 100 if old != 0 else 0 for old, new in zip(values_old, values_new)]
            common_values = [min(old, new) for old, new in zip(values_old, values_new)]
            diff_values = [abs(new - old) for old, new in zip(values_old, values_new)]

            rects_common = self.plot_common_values(common_values, i, self.colors[i])
            self.plot_difference_values(diff_values, common_values, i, self.colors[i])
            self.plot_positive_diff(
                [max(new - old, 0) for old, new in zip(values_old, values_new)], common_values, i, self.colors[i])

            self.add_annotations_and_text(rects_common, diffs, values_old, values_new)
            self.set_axes_labels()
            self.add_legend()
            self.set_titles()

    def add_legend(self):
        handles = [plt.Rectangle((0, 0), 1, 1, color=color) for color in self.colors]
        self.ax.legend(handles, self.legend_labels)



from pyvis.network import Network

from efootprint.utils.graph_tools import WIDTH, HEIGHT, set_string_max_width

COLOR_MAP = {
    "ServerBase": "red",
    "Device": "red",
    "Storage": "red",
    "UsagePattern": "blue",
    "UsageJourney": "dodgerblue",
    "UsageJourneyStep": "deepskyblue",
    "EdgeDevice": "red",
    "EdgeComponent": "red",
    "EdgeUsagePattern": "blue",
    "EdgeUsageJourney": "dodgerblue",
    "EdgeFunction": "deepskyblue",
    "JobBase": "palegoldenrod",
    "RecurrentEdgeDeviceNeed": "palegoldenrod",
    "RecurrentServerNeed": "palegoldenrod",
    "RecurrentEdgeComponentNeed": "lightgoldenrodyellow",
    "Service": "orange",
}

USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE = [
    "System", "Network", "Device", "Country", "JobBase", "RecurrentEdgeDeviceNeed",
    "Storage", "EdgeComponent", "Service", "RecurrentEdgeComponentNeed"]
INFRA_VIEW_CLASSES_TO_IGNORE = [
    "UsagePattern", "EdgeUsagePattern", "Network", "Device", "System", "UsageJourneyStep", "EdgeFunction", "Country",
    "EdgeComponent", "RecurrentEdgeComponentNeed"]


def is_included_in_graph(input_mod_obj, classes_to_ignore):
    from efootprint.all_classes_in_order import ALL_CANONICAL_CLASSES_DICT
    for ignored_class in classes_to_ignore:
        if isinstance(input_mod_obj, ALL_CANONICAL_CLASSES_DICT[ignored_class]):
            return False
    return True


def get_node_color(input_mod_obj):
    from efootprint.all_classes_in_order import ALL_CANONICAL_CLASSES_DICT
    for class_name, color in COLOR_MAP.items():
        if isinstance(input_mod_obj, ALL_CANONICAL_CLASSES_DICT[class_name]):
            return color
    return "gray"


def build_object_relationships_graph(
        input_mod_obj, input_graph=None, visited_python_ids=None, classes_to_ignore=None, width=WIDTH, height=HEIGHT,
        notebook=False):
    cdn_resources = "local"
    if notebook:
        cdn_resources = "in_line"
    if classes_to_ignore is None:
        classes_to_ignore = ["System"]
    if input_graph is None:
        input_graph = Network(notebook=notebook, width=width, height=height, cdn_resources=cdn_resources)
    if visited_python_ids is None:
        visited_python_ids = set()

    if id(input_mod_obj) in visited_python_ids:
        return input_graph
    # Python ids are used to track visited objects because there can be multiple ContextualModelingObjectAttributes
    # pointing to the same ModelingObject instance, and their e-footprint ids will be the same in this case, whereas
    # the Python ids will always be different.
    visited_python_ids.add(id(input_mod_obj))
    included_in_graph = is_included_in_graph(input_mod_obj, classes_to_ignore)
    if included_in_graph:
        input_graph.add_node(
            input_mod_obj.id, label=set_string_max_width(f"{input_mod_obj.name}", 20),
            title=set_string_max_width(str(input_mod_obj), 80),
            color=get_node_color(input_mod_obj))

    for mod_obj_attribute in input_mod_obj.mod_obj_attributes:
        if is_included_in_graph(mod_obj_attribute, classes_to_ignore):
            input_graph.add_node(
                mod_obj_attribute.id, label=set_string_max_width(f"{mod_obj_attribute.name}", 20),
                title=set_string_max_width(str(mod_obj_attribute), 80),
                color=get_node_color(mod_obj_attribute))
            if included_in_graph:
                input_graph.add_edge(input_mod_obj.id, mod_obj_attribute.id)
            else:
                recursively_create_link_with_latest_non_ignored_node(
                    input_mod_obj, mod_obj_attribute, input_graph, classes_to_ignore)

        if mod_obj_attribute not in visited_python_ids:
            build_object_relationships_graph(mod_obj_attribute, input_graph, visited_python_ids, classes_to_ignore, width, height)

    return input_graph


def recursively_create_link_with_latest_non_ignored_node(source_obj, new_obj_to_link, input_graph, classes_to_ignore):
    for mod_obj in source_obj.modeling_obj_containers:
        if is_included_in_graph(mod_obj, classes_to_ignore):
            if mod_obj.id != new_obj_to_link.id and mod_obj.id in input_graph.get_nodes():
                input_graph.add_edge(mod_obj.id, new_obj_to_link.id)
        else:
            recursively_create_link_with_latest_non_ignored_node(
                mod_obj, new_obj_to_link, input_graph, classes_to_ignore)



from typing import TYPE_CHECKING, List, Optional

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_component import EdgeComponent

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_function import EdgeFunction
    from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.hardware.edge.edge_device import EdgeDevice


class InvalidComponentNeedUnitError(Exception):
    def __init__(self, component_name: str, need_unit, expected_units: List):
        message = (
            f"RecurrentEdgeComponentNeed linked to {component_name} has incompatible unit '{need_unit}'. "
            f"Expected one of: {[str(unit) for unit in expected_units]}")
        super().__init__(message)


class WorkloadOutOfBoundsError(Exception):
    def __init__(self, workload_name: str, min_value: float, max_value: float):
        message = (
            f"Workload '{workload_name}' has values outside the valid range [0, 1]. "
            f"Found values between {min_value:.3f} and {max_value:.3f}. "
            f"Workload values must represent a percentage between 0 and 1 (0% to 100%).")
        super().__init__(message)


class RecurrentEdgeComponentNeed(ModelingObject):
    def __init__(self, name: str, edge_component: EdgeComponent, recurrent_need: ExplainableRecurrentQuantities):
        super().__init__(name)
        self.edge_component = edge_component
        self.recurrent_need = recurrent_need.set_label(f"{self.name} recurrent need")
        self.validated_recurrent_need = EmptyExplainableObject()
        self.unitary_hourly_need_per_usage_pattern = ExplainableObjectDict()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[EdgeComponent]:
        return [self.edge_component]

    @property
    def calculated_attributes(self):
        return ["validated_recurrent_need", "unitary_hourly_need_per_usage_pattern"]

    @property
    def recurrent_edge_device_needs(self):
        return self.modeling_obj_containers

    @property
    def edge_device(self) -> Optional["EdgeDevice"]:
        if not self.recurrent_edge_device_needs:
            return None
        return self.recurrent_edge_device_needs[0].edge_device

    @property
    def edge_functions(self) -> List["EdgeFunction"]:
        return list(set(sum([need.edge_functions for need in self.recurrent_edge_device_needs], start=[])))

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return list(set(sum([ef.edge_usage_journeys for ef in self.edge_functions], start=[])))

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([euj.edge_usage_patterns for euj in self.edge_usage_journeys], start=[])))

    @staticmethod
    def assert_recurrent_workload_is_between_0_and_1(
            recurrent_workload: ExplainableRecurrentQuantities, workload_name: str):
        # Convert to concurrent (or dimensionless-like unit) to get raw magnitude
        workload_magnitude = recurrent_workload.value.to(u.concurrent).magnitude
        min_value = float(workload_magnitude.min())
        max_value = float(workload_magnitude.max())

        if min_value < 0 or max_value > 1:
            raise WorkloadOutOfBoundsError(workload_name, min_value, max_value)

    def update_validated_recurrent_need(self):
        """Validate that the recurrent_need unit is compatible with the edge_component."""
        root_need_unit = self.recurrent_need.value.to_root_units().units
        expected_units = self.edge_component.compatible_root_units

        if not root_need_unit in expected_units:
            raise InvalidComponentNeedUnitError(self.edge_component.name, root_need_unit, expected_units)

        if expected_units == ["concurrent"]:
            # For dimensionless needs (like workload), ensure values are between 0 and 1
            self.assert_recurrent_workload_is_between_0_and_1(self.recurrent_need, self.name)

        self.validated_recurrent_need = self.recurrent_need.copy().set_label(
            f"Validated recurrent need of {self.name}")

    def update_dict_element_in_unitary_hourly_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_need = self.recurrent_need.generate_hourly_quantities_over_timespan(
            usage_pattern.nb_edge_usage_journeys_in_parallel, usage_pattern.country.timezone)
        self.unitary_hourly_need_per_usage_pattern[usage_pattern] = unitary_hourly_need.set_label(
            f"{self.name} unitary hourly need for {usage_pattern.name}")

    def update_unitary_hourly_need_per_usage_pattern(self):
        self.unitary_hourly_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_need_per_usage_pattern(usage_pattern)



from typing import List

from ecologits.model_repository import ModelRepository
from ecologits.utils.range_value import RangeValue

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.abstract_modeling_classes.source_objects import Source, SourceValue, Sources, SourceObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.builders.services.service_job_base_class import ServiceJob
from efootprint.core.hardware.gpu_server import GPUServer
from efootprint.builders.services.service_base_class import Service
from efootprint.constants.units import u

models = ModelRepository.from_json()

ecologits_source = Source("Ecologits", "https://github.com/genai-impact/ecologits")


class GenAIModel(Service):
    default_values =  {
            "provider": SourceObject("mistralai"),
            "model_name": SourceObject("open-mistral-7b"),
            "nb_of_bits_per_parameter": SourceValue(16 * u.dimensionless, source=Sources.HYPOTHESIS),
            "gpu_latency_alpha": SourceValue(8.02e-4 * u.ns, source=ecologits_source),
            "gpu_latency_beta": SourceValue(2.23e-2 * u.s, source=ecologits_source),
            "llm_memory_factor": SourceValue(1.2 * u.dimensionless, source=ecologits_source),
            "bits_per_token": SourceValue(24 * u.dimensionless)
            }

    sorted_provider_names = sorted(list(set([model.provider.name for model in models.list_models()])))
    list_values = {"provider": [SourceObject(provider_name) for provider_name in sorted_provider_names]}

    @staticmethod
    def generate_conditional_list_values(list_values):
        values = {}
        for provider in list_values["provider"]:
            values[provider] = [SourceObject(model.name) for model in models.list_models()
                                if model.provider.name == provider.value]

        return {"model_name": {"depends_on": "provider", "conditional_list_values": values}}

    conditional_list_values = generate_conditional_list_values(list_values)

    def __init__(self, name: str, provider: ExplainableObject, model_name: ExplainableObject, server: GPUServer,
                 nb_of_bits_per_parameter: ExplainableQuantity, llm_memory_factor: ExplainableQuantity,
                 gpu_latency_alpha: ExplainableQuantity, gpu_latency_beta: ExplainableQuantity,
                 bits_per_token: ExplainableQuantity):
        super().__init__(name=name, server=server)
        self.provider = provider.set_label(str(provider))
        self.model_name = model_name.set_label(f"{provider} model used")
        self.nb_of_bits_per_parameter = nb_of_bits_per_parameter.set_label(f"{self.name} nb of bits per parameter")
        self.llm_memory_factor = (llm_memory_factor.set_label
                                  (f"{self.name} ratio between GPU memory footprint and model size"))
        self.gpu_latency_alpha = gpu_latency_alpha.set_label("GPU latency per active parameter and output token")
        self.gpu_latency_beta = gpu_latency_beta.set_label("Base GPU latency per output_token")
        self.bits_per_token = bits_per_token.set_label(f"Number of bits per token")
        self.active_params = EmptyExplainableObject()
        self.total_params = EmptyExplainableObject()

    @property
    def calculated_attributes(self) -> List[str]:
        return ["active_params", "total_params", "base_ram_consumption"]

    def update_active_params(self):
        model = models.find_model(provider=self.provider.value, model_name=self.model_name.value)
        if isinstance(model.architecture.parameters, int) or isinstance(model.architecture.parameters, float):
            model_active_params_in_billion = model.architecture.parameters
        elif isinstance(model.architecture.parameters, RangeValue):
            model_active_params_in_billion = (model.architecture.parameters.min + model.architecture.parameters.max) / 2
        elif isinstance(model.architecture.parameters.active, RangeValue):
            model_active_params_in_billion = (model.architecture.parameters.active.min
                                              + model.architecture.parameters.active.max) / 2
        else:
            model_active_params_in_billion = model.architecture.parameters.active

        self.active_params = ExplainableQuantity(
            model_active_params_in_billion * 1e9 * u.dimensionless,
            f"{self.model_name} from {self.provider} nb of active parameters",
            left_parent=self.provider, right_parent=self.model_name, operator="query EcoLogits data with",
            source=ecologits_source)

    def update_total_params(self):
        model = models.find_model(provider=self.provider.value, model_name=self.model_name.value)
        if isinstance(model.architecture.parameters, int) or isinstance(model.architecture.parameters, float):
            model_total_params_in_billion = model.architecture.parameters
        elif isinstance(model.architecture.parameters, RangeValue):
            model_total_params_in_billion = (model.architecture.parameters.min + model.architecture.parameters.max) / 2
        elif isinstance(model.architecture.parameters.total, RangeValue):
            model_total_params_in_billion = (model.architecture.parameters.total.min
                                              + model.architecture.parameters.total.max) / 2
        else:
            model_total_params_in_billion = model.architecture.parameters.total

        self.total_params = ExplainableQuantity(
            model_total_params_in_billion * 1e9 * u.dimensionless,
            f"{self.model_name} from {self.provider} total nb of parameters",
            left_parent=self.provider, right_parent=self.model_name, operator="query EcoLogits data with",
            source=ecologits_source)

    def update_base_ram_consumption(self):
        self.base_ram_consumption = (
                self.llm_memory_factor * self.total_params * self.nb_of_bits_per_parameter).to(u.GB).set_label(
            f"{self.name} base RAM consumption")



class GenAIJob(ServiceJob):
    default_values =  {
            "output_token_count": SourceValue(1000 * u.dimensionless)
        }

    def __init__(self, name: str, service: GenAIModel, output_token_count: ExplainableQuantity):
        """
        Create a Job object requesting the Gen AI model.

        Args:
            service (GenAIModel): The Gen AI model to use for the job.
            output_token_count (SourceValue): The number of output tokens for the job.
        Returns:
            Job: An object that represents the resources needed for the task.
        """
        super().__init__(
            name or f"request to {service.model_name} installed on {service.server.name}",
            service,
            data_transferred=SourceValue(0 * u.kB),
            data_stored=SourceValue(0 * u.kB),
            request_duration=SourceValue(0 * u.s),
            compute_needed=SourceValue(0 * u.gpu),
            ram_needed=SourceValue(0 * u.GB))
        self.output_token_count = output_token_count.set_label(f"{self.name} output token count")
        self.output_token_weights = EmptyExplainableObject()

    @property
    def calculated_attributes(self) -> List[str]:
        return (["output_token_weights", "data_stored", "data_transferred", "request_duration", "ram_needed",
                 "compute_needed"] + super().calculated_attributes)

    def update_output_token_weights(self):
        self.output_token_weights = (self.output_token_count * self.service.bits_per_token).to(u.kB).set_label(
            f"{self.name} output token weights")

    def update_data_stored(self):
        self.data_stored = (SourceValue(100 * u.kB) + self.output_token_weights).set_label(f"{self.name} data stored")

    def update_data_transferred(self):
        self.data_transferred = (SourceValue(100 * u.kB) + self.output_token_weights).set_label(
            f"{self.name} data transferred")

    def update_request_duration(self):
        gpu_latency = self.output_token_count * (
            self.service.gpu_latency_alpha * self.service.active_params + self.service.gpu_latency_beta)
        self.request_duration = gpu_latency.set_label(f"{self.name} request duration")

    def update_ram_needed(self):
        self.ram_needed = SourceValue(0 * u.GB).set_label(
            f"No additional GPU RAM needed because model is already loaded in memory")

    def update_compute_needed(self):
        self.compute_needed = (
            (self.service.llm_memory_factor * self.service.active_params * self.service.nb_of_bits_per_parameter
             / self.server.ram_per_gpu)).to(u.gpu).set_label(f"{self.name} nb of required GPUs during inference")



from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.core.hardware.edge.edge_workload_component import EdgeWorkloadComponent


class EdgeApplianceComponent(EdgeWorkloadComponent):
    def __init__(self, name: str):
        super().__init__(
            name=name,
            carbon_footprint_fabrication=SourceValue(0 * u.kg),
            power=SourceValue(1 * u.W),
            lifespan=SourceValue(1 * u.year),
            idle_power=SourceValue(0 * u.W))

    @property
    def calculated_attributes(self):
        return ["power", "idle_power", "lifespan"] + super().calculated_attributes

    def update_power(self):
        edge_device = self.edge_device
        if edge_device:
            self.power = edge_device.power.copy().set_label(f"Power of {self.name}")
        else:
            self.power = EmptyExplainableObject()

    def update_idle_power(self):
        edge_device = self.edge_device
        if edge_device:
            self.idle_power = edge_device.idle_power.copy().set_label(f"Idle power of {self.name}")
        else:
            self.idle_power = EmptyExplainableObject()

    def update_lifespan(self):
        edge_device = self.edge_device
        if edge_device:
            self.lifespan = edge_device.lifespan.copy().set_label(f"Lifespan of {self.name}")
        else:
            self.lifespan = EmptyExplainableObject()


class EdgeAppliance(EdgeDevice):
    default_values = {
        "carbon_footprint_fabrication": SourceValue(100 * u.kg),
        "power": SourceValue(50 * u.W),
        "lifespan": SourceValue(6 * u.year),
        "idle_power": SourceValue(5 * u.W),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity):
        super().__init__(
            name=name,
            structure_carbon_footprint_fabrication=SourceValue(0 * u.kg),
            components=[],
            lifespan=lifespan)
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(
            f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")

    @property
    def calculated_attributes(self):
        return ["structure_carbon_footprint_fabrication"] + super().calculated_attributes

    def update_structure_carbon_footprint_fabrication(self):
        self.structure_carbon_footprint_fabrication = self.carbon_footprint_fabrication.copy().set_label(
            f"Structure fabrication carbon footprint of {self.name}")

    def after_init(self):
        if not self.components:
            appliance_component = EdgeApplianceComponent(name=f"{self.name} appliance")
            self.components = [appliance_component]
        super().after_init()

    @property
    def appliance_component(self) -> EdgeApplianceComponent:
        return self.components[0]

    @property
    def unitary_hourly_workload_per_usage_pattern(self):
        return self.appliance_component.unitary_hourly_workload_per_usage_pattern

    def self_delete(self):
        components = list(self.components)
        super().self_delete()
        for component in components:
            component.self_delete()



from typing import List

import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.abstract_modeling_classes.source_objects import SourceRecurrentValues
from efootprint.constants.units import u
from efootprint.builders.hardware.edge.edge_appliance import EdgeAppliance
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed


class RecurrentEdgeWorkloadNeed(RecurrentEdgeComponentNeed):
    def __init__(self, name: str, edge_component: EdgeComponent):
        from efootprint.abstract_modeling_classes.source_objects import SourceRecurrentValues
        super().__init__(
            name=name,
            edge_component=edge_component,
            recurrent_need=SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.concurrent)))

    @property
    def calculated_attributes(self):
        return ["recurrent_need"] + super().calculated_attributes

    def update_recurrent_need(self):
        if not self.recurrent_edge_device_needs:
            self.recurrent_need = SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.concurrent))
            return
        recurrent_edge_device_need = self.recurrent_edge_device_needs[0]
        self.recurrent_need = recurrent_edge_device_need.recurrent_workload.copy().set_label(
            f"{self.name} recurrent need")


class RecurrentEdgeWorkload(RecurrentEdgeDeviceNeed):
    default_values = {
        "recurrent_workload": SourceRecurrentValues(Quantity(np.array([1] * 168, dtype=np.float32), u.concurrent)),
    }

    def __init__(self, name: str, edge_device: EdgeAppliance, recurrent_workload: ExplainableRecurrentQuantities):
        super().__init__(
            name=name,
            edge_device=edge_device,
            recurrent_edge_component_needs=[])
        self.recurrent_workload = recurrent_workload.set_label(f"Recurrent workload for {self.name}")

    def after_init(self):
        if not self.recurrent_edge_component_needs:
            workload_need = RecurrentEdgeWorkloadNeed(
                name=f"{self.name} workload need",
                edge_component=self.edge_device.appliance_component)
            self.recurrent_edge_component_needs = [workload_need]
        super().after_init()

    @property
    def workload_need(self) -> RecurrentEdgeWorkloadNeed:
        return self.recurrent_edge_component_needs[0]

    @property
    def attribute_update_entanglements(self):
        return {
            "edge_device": self.generate_component_needs_changes_from_device_change,
        }

    def generate_component_needs_changes_from_device_change(self, change: List[EdgeAppliance]):
        old_edge_appliance, new_edge_appliance = change[0], change[1]
        component_needs_changes = [
            [self.workload_need.edge_component, new_edge_appliance.appliance_component],
        ]
        return component_needs_changes

    @property
    def unitary_hourly_workload_per_usage_pattern(self):
        return self.workload_need.unitary_hourly_need_per_usage_pattern



from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.usage.job import JobBase

if TYPE_CHECKING:
    from efootprint.core.usage.usage_journey import UsageJourney
    from efootprint.core.usage.usage_pattern import UsagePattern
    from efootprint.core.hardware.network import Network


class UsageJourneyStep(ModelingObject):
    default_values =  {"user_time_spent": SourceValue(1 * u.min)}

    def __init__(self, name: str, user_time_spent: ExplainableQuantity, jobs: List[JobBase]):
        super().__init__(name)
        self.user_time_spent = user_time_spent
        self.user_time_spent.set_label(f"Time spent on step {self.name}")
        self.jobs = jobs

    @property
    def usage_journeys(self) -> List["UsageJourney"]:
        return self.modeling_obj_containers

    @property
    def usage_patterns(self) -> List["UsagePattern"]:
        return list(set(sum([uj.usage_patterns for uj in self.usage_journeys], start=[])))

    @property
    def networks(self) -> List["Network"]:
        return list(set([up.network for up in self.usage_patterns]))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List["UsageJourney"] | List[JobBase]:
        return self.jobs + self.networks
